@@ -6,8 +6,8 @@ def get_output_error_code(cmd):
     """Get stdout, stderr, and exit code from running a command"""
     p = Popen(cmd, stdout=PIPE, stderr=PIPE)
     out, err = p.communicate()
-    out = out.decode("utf8", "replace")
-    err = err.decode("utf8", "replace")
+    out = out.decode("utf8", "replace")  # type:ignore
+    err = err.decode("utf8", "replace")  # type:ignore
     return out, err, p.returncode
 
 
